package meerkat.mixer.main;

import com.google.protobuf.ByteString;
import meerkat.crypto.mixnet.MixerOutput;
import meerkat.protobuf.BulletinBoardAPI;
import meerkat.protobuf.Crypto;
import meerkat.protobuf.Mixing;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.List;

/**
 * Created by Tzlil on 12/17/2015.
 * provide convert operation from batch data to meerkat.mixer.mixing output and backwards
 */
public class BatchConverter {

    private final int n,layers;

    /**
     * constructor
     * @param n
     * @param layers
     */
    public BatchConverter(int n,int layers){
        this.n = n;
        this.layers = layers;
    }

    /**
     * convert integer to byte string
     * @param a
     * @return a as byte string
     */
    private ByteString Integer2ByteString(int a){
        return ByteString.copyFrom(BigInteger.valueOf(a).toByteArray());
    }

    /**
     * convert byte string to integer
     * @param bs
     * @return bs as int
     */
    private int ByteString2Integer(ByteString bs) {
        return Integer.valueOf(bs.toString());
    }

    /**
     * convert meerkat.mixer.mixing output to batch data
     * @param mixerOutput
     * @return meerkat.mixer.mixing output as list of batch data
     */
    public List<BulletinBoardAPI.BatchData> MixerOutput2BatchData(MixerOutput mixerOutput) {

        List<BulletinBoardAPI.BatchData> result = new ArrayList<BulletinBoardAPI.BatchData>();

        result.add(BulletinBoardAPI.BatchData.newBuilder()
                .setData(Integer2ByteString(n))
                .build());

        for (Mixing.ZeroKnowledgeProof[] zkpLayer : mixerOutput.getProofs()) {
            for (Mixing.ZeroKnowledgeProof zkp : zkpLayer) {
                result.add(BulletinBoardAPI.BatchData.newBuilder()
                        .setData(zkp.toByteString())
                        .build());
            }
        }
        for (Crypto.RerandomizableEncryptedMessage[] encryptionLayer : mixerOutput.getEncryptedMessages()) {
            for (Crypto.RerandomizableEncryptedMessage encryption : encryptionLayer) {
                result.add(BulletinBoardAPI.BatchData.newBuilder()
                        .setData(encryption.toByteString())
                        .build());
            }
        }
        return result;
    }

    /**
     * convert batch data list to meerkat.mixer.mixing output
     * @param batchDataList
     * @return batch data list as MixerOutput
     * @throws Exception
     */
    public MixerOutput BatchDataList2MixerOutput
    (List<BulletinBoardAPI.BatchData> batchDataList) throws Exception {

        if (n != ByteString2Integer(batchDataList.remove(0).getData())){
            throw new Exception();
        }

        int nDiv2 = n >>1;
        Mixing.ZeroKnowledgeProof[][] proofs = new Mixing.ZeroKnowledgeProof[layers][nDiv2];
        for (int layer = 0; layer < layers; layer++)
        {
            for (int proofIndex = 0 ; proofIndex < nDiv2 ; proofIndex ++)
            {
                proofs[layer][proofIndex] = Mixing.ZeroKnowledgeProof.parseFrom(batchDataList.remove(0).getData());
            }
        }

        Crypto.RerandomizableEncryptedMessage[][] encryptions
                = new Crypto.RerandomizableEncryptedMessage[layers + 1][n];
        for (int layer = 0; layer < layers + 1; layer++)
        {
            for (int encryptionIndex = 0 ; encryptionIndex < n ; encryptionIndex ++)
            {
                encryptions[layer][encryptionIndex] = Crypto.RerandomizableEncryptedMessage
                        .parseFrom(batchDataList.remove(0).getData());
            }
        }

        return new meerkat.mixer.mixing.MixerOutput(n,layers,proofs,encryptions);

    }

}

